#include "alsa.h"

typedef struct {
	snd_pcm_t            * handle;
	snd_pcm_hw_params_t  * params;
} alsa_device_t;

typedef struct {
	alsa_device_t device;
	wios_buffer_t buffer;
} alsa_priv_t;

static const snd_pcm_format_t alsa_format[] = {
	SND_PCM_FORMAT_U8,
	SND_PCM_FORMAT_S16_LE,
	SND_PCM_FORMAT_S24_LE,
	SND_PCM_FORMAT_S32_LE,
	SND_PCM_FORMAT_UNKNOWN
};



/* STATIC FUNCTION PROTOTYPE */
/* -------------------------------------------------- */

static int priv_init(wios_file_t * ft);
static int device_write(wios_file_t * ft, wios_buffer_t * buf, int len);
static int device_read(wios_file_t * ft, wios_buffer_t * buf, int len);

static int alsa_device_recover(wios_file_t * ft, int len, int err);
static char * alsa_buffer_malloc(wios_file_t * ft, wios_buffer_t * buf);



/* HANDLER FUNCTION */
/* -------------------------------------------------- */

/* --------------------------------------------------
 * ALSA
 * 同期時のデバイス設定について
 * 
 * 再録同期時にはDAとADのフォーマットは異なる値をデバイスに設定可能。
 * 具体的には
 * サンプリングレート、チャンネル数、エンコーディング
 * を異なるものに設定することができる。
 * -------------------------------------------------- */

#define SET(func,args,fmt,...) do { if ((err = func args) < 0) {wios_fail("ERROR: %s: " fmt " (%s) (%s)", #func,__VA_ARGS__, snd_strerror(err), ft->file_name); goto error;} } while (0)

static int device_init(wios_file_t * ft)
{
	int err, dir = 0;
	int upp; /* usec per period */
	alsa_priv_t    * priv = ft->priv;
	alsa_device_t  * dev = &priv->device;
	wios_signal_t  * sig = &ft->signal;
	wios_fmtinfo_t * fmt = &ft->fmtinfo;

	/* PCMデバイスをPLAYBACK か CAPTUREモードで開く */
	SET(snd_pcm_open,
		(&dev->handle, ft->dev_name, (ft->io_type == wios_da ? SND_PCM_STREAM_PLAYBACK : SND_PCM_STREAM_CAPTURE), dir),
		"cannot open audio device `%s'", ft->dev_name);
	
	/* ハードウェアパラメータの領域を確保 */
	SET(snd_pcm_hw_params_malloc, (&dev->params),
		"cannot allocate hardware parameter strucutre", NULL);
	/* ハードウェアパラメータを初期化 */
	SET(snd_pcm_hw_params_any, (dev->handle, dev->params),
		"cannot initialize hardware parameter strucutre", NULL);

	/* アクセスタイプをRW_INTERLEAVEDに設定 */
	SET(snd_pcm_hw_params_set_access, (dev->handle, dev->params, SND_PCM_ACCESS_RW_INTERLEAVED),
		"cannot set access type RW_INTERLEAVED", NULL);
	/* 量子化ビット数を設定 */
	SET(snd_pcm_hw_params_set_format, (dev->handle, dev->params, alsa_format[fmt->format]),
		"cannot set format `%s'", wios_format_to_string(fmt->format));
	/* サンプリングレートを設定 */
	SET(snd_pcm_hw_params_set_rate, (dev->handle, dev->params, fmt->rate, dir),
		"cannot set sampling rate `%d'", fmt->rate);
	/* チャンネル数を設定 */
	SET(snd_pcm_hw_params_set_channels, (dev->handle, dev->params, fmt->channels),
		"cannot set channel count `%d'", fmt->channels);
	/* リングバッファのperiod数を指定 */
	SET(snd_pcm_hw_params_set_periods, (dev->handle, dev->params, sig->execinfo.periods, dir),
		"cannot set period count `%d'", sig->execinfo.periods);

	/* リングバッファサイズをフレーム数で指定 */
  int request_buffer_size = sig->execinfo.periodsize * sig->execinfo.periods;
  int returned_buffer_size = request_buffer_size;
	SET(snd_pcm_hw_params_set_buffer_size_near,
      (dev->handle, dev->params, (snd_pcm_uframes_t *)&returned_buffer_size),
		  "cannot set ring buffer size near `%d'", request_buffer_size);
  if (returned_buffer_size != request_buffer_size) {
    int old_periodsize = sig->execinfo.periodsize;
    sig->execinfo.periodsize = returned_buffer_size / sig->execinfo.periods;
    printf("ring buffer size is set to %d for `%d'. ",
           returned_buffer_size, request_buffer_size);
    printf("periodsize is changed to %d from `%d' (%s)\n",
           sig->execinfo.periodsize, old_periodsize, ft->file_name);
  }

	/* ハードウェアパラメータをPCMデバイスに適用 */
	SET(snd_pcm_hw_params, (dev->handle, dev->params),
		"cannot install PCM hardware configuration", NULL);

	/* usec per period 取得 */
	SET(snd_pcm_hw_params_get_period_time, (dev->params, &upp, &dir),
		"cannot get period time", NULL);

	ft->dp = dev->handle;
	sig->sec_per_frame = CHANGE_USEC_TO_SEC(upp) / (double)sig->execinfo.periodsize;

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static wios_file_t * open_read(
	const char             * path,
	const char             * dev,
	const wios_execinfo_t  * info)
{
	return open_read_(path, dev, info, priv_init);
}

static wios_file_t * open_write(
		const char             * path,
		const char             * dev,
		const wios_execinfo_t  * info,
		const wios_fmtinfo_t   * fmt)
{
	return open_write_(path, dev, info, fmt, priv_init);
}

static void * playback(void * arg)
{
	/* 再生準備が終わるまで呼び出し元をロック */
	pthread_mutex_lock(&ready_mutex);
	wios_file_t    * ft = (wios_file_t *)arg;
	alsa_priv_t    * priv = ft->priv;
	wios_buffer_t  * buf = &priv->buffer;

	int frames;
	int done = 0;
	int len = ft->signal.execinfo.periodsize;
	int req_nframes;

	if (ft->signal.execinfo.sec == WIOS_ALL_SEC)
		req_nframes = ft->signal.all_frames;
	else
		req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;

	if (alsa_buffer_malloc(ft, buf) == NULL)
		goto error;
 
	/* 1periodsize分のデータを先に読み込んでおく */
	frames = lsf_file_read(ft, buf->fl, len);

	/* 再生準備完了をメインスレッドに知らせる */
	pthread_cond_signal(&ready_cond);
	pthread_mutex_unlock(&ready_mutex);

	if (wios_mode == wios_sync) {
		/* メインスレッドで同期を開始させるまで待機 */
		pthread_mutex_lock(&sync_mutex);
		while (!sync_start_flag)
			pthread_cond_wait(&sync_cond, &sync_mutex);
		pthread_mutex_unlock(&sync_mutex);
	}

	/* 以下同期スタート */
	if (print_mesg_flag)
		fprintf(stdout, "Start playback (%s)\n", ft->file_name);
  if (print_time_flag)
    fprintf(stdout, "Playback_start_time:%lf\n", wios_get_time());

	/* 再生ループ */
	while (1) {
		/* SIGINTを受け取っていたらブレーク */
		if (loop_break_flag) break;

		device_write(ft, buf, frames);

		done += frames;

		if (done >= req_nframes) break;

		len = min(len, req_nframes-done);

		frames = lsf_file_read(ft, buf->fl, len);

		/* ファイル読み取りエラー */
		if (frames <= 0) break;
	}

	/* 再生終了を待つ */
	snd_pcm_drain(ft->dp);
	/* 再生終了を録音スレッドに知らせる */
	playback_end_flag = WIOS_TRUE;

	if (print_mesg_flag)
		fprintf(stdout, "End playback (%s)\n", ft->file_name);

	wios_buffer_free(buf);

	return NULL;

error:
	if (pthread_mutex_destroy(&ready_mutex) == EBUSY)
		pthread_mutex_unlock(&ready_mutex);

	wios_buffer_free(buf);

	return NULL;
}

void * capture(void * arg)
{
	/* 録音準備が終わるまで呼び出し元をロック */
	pthread_mutex_lock(&ready_mutex);
	wios_file_t * ft = (wios_file_t *)arg;
	alsa_priv_t * priv = ft->priv;
	wios_buffer_t * buf = &priv->buffer;

	int frames;
	int done = 0;
	int len = ft->signal.execinfo.periodsize;
	int req_nframes;

	if (wios_mode == wios_sync) {
		if (ft->signal.execinfo.sec == WIOS_ALL_SEC)
			req_nframes = ftda->signal.all_frames;
		else
			req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;
	} else {
		req_nframes = ft->fmtinfo.rate * ft->signal.execinfo.sec;
	}

	if (alsa_buffer_malloc(ft, buf) == NULL)
		goto error;

	/* 録音準備完了をメインスレッドに知らせる */
	pthread_cond_signal(&ready_cond);
	pthread_mutex_unlock(&ready_mutex);

	if (wios_mode == wios_sync) {
		/* メインスレッドで同期を開始させるまで待機 */
		pthread_mutex_lock(&sync_mutex);
		while (!sync_start_flag)
			pthread_cond_wait(&sync_cond, &sync_mutex);
		pthread_mutex_unlock(&sync_mutex);
	}

	/* 以下同期スタート */
	if (print_mesg_flag)
		fprintf(stdout, "Start capture (%s)\n", ft->file_name);
  if (print_time_flag)
    fprintf(stdout, "Capture_start_Time:%lf\n", wios_get_time());

	/* 録音ループ */
	while (1) {
		/* SIGINTを受け取っていたらブレーク */
		if (loop_break_flag) break;

		if (wios_mode != wios_rec || ft->signal.execinfo.sec != WIOS_ALL_SEC)
			len = min(len, req_nframes-done);

		frames = device_read(ft, buf, len);

		if (wios_mode != wios_rec || ft->signal.execinfo.sec != WIOS_ALL_SEC)
			frames = min(frames, req_nframes-done);

		lsf_file_write(ft, buf->fl, frames);

		done += frames;

		/* 録音モードで秒数がしていされなかった場合はブレークしない */
		if (wios_mode != wios_rec || ft->signal.execinfo.sec != WIOS_ALL_SEC)
			if (done >= req_nframes) break;
	}

	if (print_mesg_flag)
		fprintf(stdout, "End capture (%s)\n", ft->file_name);

	wios_buffer_free(buf);

	return NULL;

error:
	if (pthread_mutex_destroy(&ready_mutex) == EBUSY)
		pthread_mutex_unlock(&ready_mutex);

	wios_buffer_free(buf);

	return NULL;
}

void finish(wios_file_t * ft)
{
	alsa_priv_t * priv = ft->priv;
	if (ft) {
		if (ft->fp) {
			sf_close(ft->fp);
			ft->fp = NULL;
		}
		if (ft->dp) {
			snd_pcm_close(ft->dp);
			ft->dp = NULL;
		}
		if (priv->device.params) {
			snd_pcm_hw_params_free(priv->device.params);
			priv->device.params = NULL;
		}
		if (priv) {
			free(priv);
			priv = NULL;
		}
		if (ft) {
			free(ft);
			ft = NULL;
		}
	}
}



/* STATIC FUNCTION */
/* -------------------------------------------------- */

static int priv_init(wios_file_t * ft)
{
	alsa_priv_t * priv;

	if ((ft->priv = malloc(sizeof(alsa_priv_t))) == NULL)
		goto error;
	priv = ft->priv;

	priv->buffer.fl = NULL;
	priv->buffer.dv = NULL;

	return WIOS_SUCCESS;

error:
	return WIOS_EOF;
}

static int device_write(wios_file_t * ft, wios_buffer_t * buf, int nframes)
{
	int i, err;
	int n = ft->signal.byte_per_item*ft->fmtinfo.channels;

	if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24)
		n = 3*ft->fmtinfo.channels;

	switch (ft->fmtinfo.format) {
		case WIOS_PCM_FORMAT_U8: {
			uint8_t * buf1 = (uint8_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf1[i] = WIOS_CONVERT_32_TO_U8BIT(buf->fl[i]);
				DA_CLIP_WARN(WIOS_U8BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S16: {
			int16_t * buf1 = (int16_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf1[i] = WIOS_CONVERT_32_TO_16BIT(buf->fl[i]);
				DA_CLIP_WARN(WIOS_16BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S24: {
			int24_t * buf1 = (int24_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf1[i] = WIOS_CONVERT_32_TO_24BIT(buf->fl[i]);
				DA_CLIP_WARN(WIOS_24BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S32: {
			int32_t * buf1 = (int32_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf1[i] = buf->fl[i];
				DA_CLIP_WARN(WIOS_32BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		default: {
			break;
		}
	}

	err = snd_pcm_writei(ft->dp, buf->dv, nframes);

	ft->success_byte_cnt += err > 0 ? err*n : 0;

	alsa_device_recover(ft, nframes, err);

	return err;
}

static int device_read(wios_file_t * ft, wios_buffer_t * buf, int nframes)
{
	int i, err;
	int n = ft->signal.byte_per_item*ft->fmtinfo.channels;

	if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24)
		n = 3*ft->fmtinfo.channels;

	err = snd_pcm_readi(ft->dp, buf->dv, nframes);

	ft->success_byte_cnt += err > 0 ? err*n : 0;

	switch (ft->fmtinfo.format) {
		case WIOS_PCM_FORMAT_U8: {
			uint8_t * buf1 = (uint8_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf->fl[i] = WIOS_CONVERT_U8_TO_32BIT(buf1[i]);
				AD_CLIP_WARN(WIOS_U8BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S16: {
			int16_t * buf1 = (int16_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf->fl[i] = WIOS_CONVERT_16_TO_32BIT(buf1[i]);
				AD_CLIP_WARN(WIOS_16BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S24: {
			int24_t * buf1 = (int24_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf->fl[i] = WIOS_CONVERT_24_TO_32BIT(buf1[i]);
				AD_CLIP_WARN(WIOS_24BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		case WIOS_PCM_FORMAT_S32: {
			int32_t * buf1 = (int32_t *)buf->dv;
			for (i = 0; i < nframes*ft->fmtinfo.channels; i++) {
				buf->fl[i] = buf1[i];
				AD_CLIP_WARN(WIOS_32BIT_CLIP_COUNT(buf1[i], ft->clips));
				WIOS_RENEW_MIN_MAX_FRAME(buf1[i], ft->min_item_val, ft->max_item_val);
			}
			break;
		}
		default: {
			break;
		}
	}

	alsa_device_recover(ft, nframes, err);

	return err;
}

/* バッファアンダーラン等が起きたときに回復を試みる */
static int alsa_device_recover(wios_file_t * ft, int nframes, int err)
{
	int n;
	char * run_type, * acs_type;

	if (ft->io_type == wios_da)
		run_type = "under", acs_type = "write";
	else
		run_type = "over" , acs_type = "read";

	if (err == -EPIPE) {
		n = nframes*ft->signal.byte_per_item*ft->fmtinfo.channels;
		DEVICE_RUN_WARN("%s-run occurred. missing data: %d bytes", run_type, n);
		if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24)
			ft->missing_byte_cnt += (n*3)/4;
		else 
			ft->missing_byte_cnt += n;
		if ((err = snd_pcm_prepare(ft->dp)) < 0)
			DEVICE_RUN_WARN("cannot recovery from %s-run, prepare failed (%s)", acs_type, snd_strerror(err));
	} 
	else if (err < 0)
		DEVICE_RUN_WARN("%si device error (%s). missing data: unknown", acs_type, snd_strerror(err));
	else if (err != nframes) {
		n = (nframes-err)*ft->signal.byte_per_item*ft->fmtinfo.channels;
		DEVICE_RUN_WARN("device short %s, %s %d / %d frames. missing data: %d bytes",
			acs_type, acs_type, err, nframes, n);
		if (ft->fmtinfo.format == WIOS_PCM_FORMAT_S24)
			ft->missing_byte_cnt += (n*3)/4;
		else 
			ft->missing_byte_cnt += n;
	}
	
	return err;
}

char * alsa_buffer_malloc(wios_file_t * ft, wios_buffer_t * buf) {
	return wios_buffer_malloc(ft, buf, ft->signal.buf_size);
}



/* GLOBAL FUNCTION */
/* -------------------------------------------------- */

int alsa_device_check(const char * dev)
{
	snd_pcm_t            * handle;
	int ret;

	ret = (snd_pcm_open(&handle, dev, SND_PCM_STREAM_PLAYBACK, 0) < 0) ? WIOS_TRUE : WIOS_FALSE;
	if (handle) snd_pcm_close(handle);

	return ret;
}

wios_format_handler_t const * wios_set_alsa_format_handler(void)
{
	static wios_format_handler_t const handler = {
		device_init,
		open_read, open_write,
		playback, capture,
		finish
	};

	return &handler;
}
